"""
    Copyright 2006,2007 Riccardo Taormina (den3b@hotmail.com)
    
    This file is part of SmartChameleons.

    SmartChameleons is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    SmartChameleons is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SmartChameleons; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""

#The MODEL
from Chameleon import Chameleon
import random
import math
from mappings import *
from RL_CON import AI
import pickle
from ActionGroup import ActionGroup

class Model:
    def __init__(self, parent, envDict, chamDicts, colorList):
        # the environment
        self.envDict = envDict
        self.chamDicts = chamDicts
        self.colorList = colorList
        self.nCycles = self.envDict["nCycles"]
        self.worldXSize = self.envDict["worldXsize"]
        self.worldYSize = self.envDict["worldYsize"]
        self.parent = parent
        # number of chams for each color
        self.breedNumbers = dict()
        i=0
        for color in self.colorList:
            self.breedNumbers[color]=self.chamDicts[color]["startNumber"]
            i = i + 1
        
        self.chList = []
        self.time = 0 #start the model time

    # objects
    def buildObjects(self):
        i = 0
        for color in self.colorList:
            for j in range(self.breedNumbers[color]):
                newCham = Chameleon(i+j,color,None,self.worldXSize,self.worldYSize)
                item = self.parent.canvas.create_image(
                    newCham.xPosition*36,newCham.yPosition*36,
                    image=self.parent.cham_images[color])
                newCham.graphItem = item #creating the graphic item (reference) for the chameleon.
                self.chList.append(newCham)
            i+=self.breedNumbers[color]

        self.breedAI = dict()

        for color in self.colorList:
            try :
                brainFile = open(self.chamDicts[color]["brainPath"], 'rb') #loading the brains for the simulation
                brain = pickle.load(brainFile)
                self.breedAI[color] = AI(brain)
            except IOError:
                self.breedAI[color] = None
            
    # actions        
    def buildActions(self):
        #defining the action 

        self.chameleonsActionGroup = ActionGroup ("chameleons")

        def moveChams(self):
            random.shuffle(self.chList)
            for aCham in self.chList:
                #if the chameleon is smart then perform a smart move. Otherwise a random move would be done
                if self.breedAI[aCham.color] != None:
                    state = self.checkState(aCham)
                    if ((state == [0]*25) and (self.chamDicts[aCham.color]["type"]=='c')):
                        aCham.move()
                    else :
                        changes = self.breedAI[aCham.color].computeSmartMove(state)
                        aCham.move(changes)
                else :
                    aCham.move()                    

              
              
        def handleCollisions(self):
            #checks collisions handling the color changes
            for cham in self.chList:
                cham.justChanged = False
            i = 0
            for cham1 in self.chList:
                if cham1.hasntJustChanged():
                    for cham2 in self.chList[i+1:len(self.chList)]:
                        if cham2.hasntJustChanged():
                            if cham1.color != cham2.color and \
                                self.distance(cham1.xPosition,cham1.yPosition,
                                        cham2.xPosition,cham2.yPosition) ==0 :
                                self.changeColors(cham1,cham2)
                i += 1

        self.chameleonsActionGroup.actions = [moveChams, handleCollisions]
        self.checkActionGroup = ActionGroup ("check")
            
        def check(self):
            #checks if the Chameleons are all of the same color
            flag = True
            color = self.chList[0].color
            for chameleon in self.chList[1:]:
                if chameleon.color != color:
                    flag = False
                    break
            self.parent.onlyOneColor = flag

        self.checkActionGroup.action = check

        self.actionGroupList = ["ask_chameleons", "check_model"]



    def step(self):
        #performs a step of the model simulation. Executes each action of the model action list.
        #this entails asking the chameleons to move then checking if they're all of the same color 
        self.time += 1
        for action_type in self.actionGroupList :
            if action_type == "ask_chameleons" :
                for action in self.chameleonsActionGroup.actions :
                    action(self)
            elif action_type == "check_model" :
                self.checkActionGroup.action(self)




    def distance(self,x1,y1,x2,y2):
        #general 2 dimension distance calculation
        x1 = (x1+self.worldXSize) % self.worldXSize
        x2 = (x2+self.worldXSize) % self.worldXSize
        y1 = (y1+self.worldYSize) % self.worldYSize
        y2 = (y2+self.worldYSize) % self.worldYSize
        return math.sqrt((x1-x2)**2+(y1-y2)**2)

    def checkState(self,cham):
        #checks the state of the smart chameleon. The smart chameleon performs a move by checking its state looking around himself,
        #2 cells away in each of the 8 directions.
        #after evaluating the state, the BRAIN would suggest the move according to the POLICY created via REINFORCEMENT LEARNING.
        #the state is identified by a 5x5 matrix where each cell is equal to 1 if one or more ENEMY chameleons are present on the
        #relative WORLD cell; it has 0 otherwise (the cell is empty or occupied by chameleons of the same color)
        state=[0]*25
        for otherCham in self.chList:
            if cham.color != otherCham.color :
                distance = self.distance(cham.xPosition,cham.yPosition,
                                         otherCham.xPosition,otherCham.yPosition)
                if distance <3:
                    position = [cham.xPosition - otherCham.xPosition,cham.yPosition - otherCham.yPosition]
                    state[positionToBox(position)]=1                        
        return state

    def changeColors(self,cham1,cham2):
        #changes the colors of the chameleons after their collision. The color of the image in the canvas is also changed.
        #FOR NOW THE CHANGING IS DECIDED BY THE 1ST CHAMELEON
        
        if self.chamDicts[cham1.color]["chgBehav"]=='r':
            rankList = self.computeRanks(cham1.color,cham2.color,reverse = False)
            changingColor = rankList[0][0]
        elif self.chamDicts[cham1.color]["chgBehav"]=='c':
            rankList = self.computeRanks(cham1.color,cham2.color, reverse = True)
            changingColor = rankList[0][0]
  
        else :
            changingColorList = self.chamDicts[cham1.color]["chgColors"][:]
            # insert cham1.color if selected from the GUI
            if self.chamDicts[cham1.color]["myCol"]=='Y':
                changingColorList.append(cham1.color)

            if cham2.color not in changingColorList:
                if self.chamDicts[cham1.color]["hitCol"]=='Y':
                    changingColorList.append(cham2.color)
            else:
                if (self.chamDicts[cham1.color]["hitCol"]=='N') and (self.chamDicts[cham1.color]["hitIfCol"]=='N'):
                    changingColorList.remove(cham2.color)
            try :
                if self.chamDicts[cham1.color]["chgProb"] == 'e':
                    changingColor = random.choice(changingColorList)
                else :
                    if self.chamDicts[cham1.color]["chgProb"] == 'c':
                        rankList = self.computeRanks(cham1.color,cham2.color,reverse = True)
                    elif self.chamDicts[cham1.color]["chgProb"] == 'r':
                        rankList = self.computeRanks(cham1.color,cham2.color,reverse = False)    
                    cumProbs = self.computeCumProbs(rankList)
                    length = len(cumProbs)
                    sel = random.random()
                    for index in range(length):
                        if sel<cumProbs[index]:
                            changingColor = rankList[index][0]
                            break
            except IndexError:
##                print 'INDEX ERROR --> cause cham1 = %s and cham2 = %s' % (cham1.color,cham2.color)
                changingColor = ''
                pass

##        print "The %s chameleon hit a %s chameleon and are changing to %s" % (cham1.color,cham2.color,changingColor)
##        print "Its behaviour is %s -- %s " % (self.chamDicts[cham1.color]["chgBehav"],self.chamDicts[cham1.color]["chgProb"])
##        print self.breedNumbers
##        print 
##        print
        if changingColor != '':    
            self.breedNumbers[cham1.color]=self.breedNumbers[cham1.color]-1;
            self.breedNumbers[cham2.color]=self.breedNumbers[cham2.color]-1;
            self.breedNumbers[changingColor]=self.breedNumbers[changingColor]+2;        
            cham1.changeColor(changingColor)
            cham2.changeColor(changingColor)
            self.parent.canvas.itemconfigure(cham1.graphItem,image=self.parent.cham_images[changingColor])
            self.parent.canvas.itemconfigure(cham2.graphItem,image=self.parent.cham_images[changingColor])
            for color in self.colorList:
                self.parent.breedNumbers[color].set(self.breedNumbers[color])

    def computeCumProbs(self,rankList):
        #based simply on ranking
##        sum_ = 0
##        for item in rankList:
##            print item[1]
##            sum_ = sum_ + item[1]        
        length=len(rankList)

        sum_=0
        probs = [0.0]*length
        for j in range(length):
            probs[length-j-1] = float(j+1)/(length+1)
            sum_ = sum_ + float(j+1)/(length+1)

        cumProbs = [0]*length
        for i in range(length):
            if i == 0:
                cumProbs[i]=probs[i]/sum_
            else:
                cumProbs[i]=probs[i]/sum_+cumProbs[i-1]
        cumProbs[length-1] = 1


        return cumProbs


    def computeRanks(self,color_1, color_2, reverse):
        from operator import itemgetter
        temp = dict(self.breedNumbers)
        if self.chamDicts[color_1]["myCol"]=='N':
            del temp[color_1]
        if self.chamDicts[color_1]["hitCol"]=='N':
            del temp[color_2]
        rankList = temp.items()
        rankList.sort(key = itemgetter(1), reverse=reverse)
        return rankList

        





